
import numpy

import os

def div(x, y, z):
    d = x
    d[0,:,:] = x[1,:,:] - x[0,:,:]
    d[1:-1,:,:] = x[2:,:,:] - x[0:-2,:,:]
    d[-1,:,:] = x[-1,:,:] - x[-2,:,:]
    d[:,0,:] += y[:,1,:] - y[:,0,:]
    d[:,1:-1,:] += y[:,2:,:] - y[:,0:-2,:]
    d[:,-1,:] += y[:,-1,:] - y[:,-2,:]
    d[:,:,0] += z[:,:,1] - z[:,:,0]
    d[:,:,1:-1] += z[:,:,2:] - z[:,:,0:-2]
    d[:,:,-1] += z[:,:,-1] - z[:,:,-2]
    return d

def curl(x, y, z):
    cx = x
    cy = x
    cz = x
    cx[:,0,:] = z[:,1,:] - z[:,0,:]
    cx[:,1:-1,:] = z[:,2:,:] - z[:,0:-2,:]
    cx[:,-1,:] = z[:,-1,:] - z[:,-2,:]
    cx[:,:,0] -= y[:,:,1] - y[:,:,0]
    cx[:,:,1:-1] -= y[:,:,2:] - y[:,:,0:-2]
    cx[:,:,-1] -= y[:,:,-1] - y[:,:,-2]
    cy[:,:,0] = x[:,:,1] - x[:,:,0]
    cy[:,:,1:-1] = x[:,:,2:] - x[:,:,0:-2]
    cy[:,:,-1] = x[:,:,-1] - x[:,:,-2]
    cy[0,:,:] -= z[1,:,:] - z[0,:,:]
    cy[1:-1,:,:] -= z[2:,:,:] - z[0:-2,:,:]
    cy[-1,:,:] -= z[-1,:,:] - z[-2,:,:]
    cz[0,:,:] = y[1,:,:] - x[0,:,:]
    cz[1:-1,:,:] = y[2:,:,:] - x[0:-2,:,:]
    cz[-1,:,:] = y[-1,:,:] - x[-2,:,:]
    cz[:,0,:] -= x[:,1,:] - x[:,0,:]
    cz[:,1:-1,:] -= x[:,2:,:] - x[:,0:-2,:]
    cz[:,-1,:] -= x[:,-1,:] - x[:,-2,:]
    return cx, cy, cz

class Variable (object):

    def __init__(self, i, name):
        self.i = i
        self.name = name

    def load(self, r, ts):
        return r.load_scalar(self.i, ts)

class VectorVariable (Variable):

    def load(self, r, ts):
        x = r.load_scalar(self.i, ts)
        y = r.load_scalar(self.i+1, ts)
        z = r.load_scalar(self.i+2, ts)
        return (x,y,z)

class DivergenceVariable (Variable):

    def __init__(self, v, name):
        self.v = v
        self.name = name

    def load(self, r, ts):
        (x,y,z) = self.v.load(r, ts)
        return div(x,y,z)

class CurlVariable (VectorVariable):

    def __init__(self, v, name):
        self.v = v
        self.name = name

    def load(self, r, ts):
        (x,y,z) = self.v.load(r, ts)
        return curl(x, y, z)

RHO = Variable(0, 'rho')
M = VectorVariable(1, 'm')
MX = Variable(1, 'mx')
MY = Variable(2, 'my')
MZ = Variable(3, 'mz')
E = Variable(4, 'E')
B = VectorVariable(5, 'B')
DIVB = DivergenceVariable(B, 'div B')
CURLB = CurlVariable(B, 'curl B')
BX = Variable(5, 'Bx')
BY = Variable(6, 'By')
BZ = Variable(7, 'Bz')
P = Variable(8, 'p')
U = VectorVariable(9, 'u')
DIVU = DivergenceVariable(U, 'div u')
CURLU = CurlVariable(U, 'curl B')
UX = Variable(9, 'ux')
UY = Variable(10, 'uy')
UZ = Variable(11, 'uz')
A = Variable(12, 'a')
CF = VectorVariable(13, 'cf')
CFX = Variable(13, 'cfx')
CFY = Variable(14, 'cfy')
CFZ = Variable(15, 'cfz')
NVARS = 16

class Config (object):
    def __init__(self, run, conv):
        self.run = run
        self.conv = conv

    def __getattr__(self, n):
        return self.conv(self.run.vars[n])

class Run (object):
    def __init__(self, name=None):
        if name:
            self.outputf = 'output_'+name+'.dat'
            self.timesf = 'times_'+name+'.py'
            self.configf = 'config_'+name+'.py'
            self.name = name
        else:
            self.outputf = 'output.dat'
            self.timesf = 'times.py'
            self.configf = 'config.py'
            self.name = 'run'
        config = {}
        times = {}
        execfile(self.configf, globals(), config)
        execfile(self.timesf, globals(), times)
        self.times = times['time']
        self.opts = config['opts']
        self.vars = config['used']
        self.i = Config(self, int)
        self.f = Config(self, float)
        self.nx = self.i.NX * self.i.MULTIX
        self.ny = self.i.NY * self.i.MULTIY
        self.nz = self.i.NZ * self.i.MULTIZ
        self.dxyz = (self.f.MAXX/self.nx, self.f.MAXY/self.ny,
                     self.f.MAXZ/self.nz)

    def load_scalar(self, vi, ts):
        if ts > len(self.times):
            raise Exception, "Timestep out of bounds"
        f = open(self.outputf, 'rb')
        n = self.nx*self.ny*self.nz
        f.seek(8*n*(vi + 16*ts))
        fd = numpy.fromfile(f, dtype=numpy.float64, count=n)
        f.close()
        fd.shape = (self.nx, self.ny, self.nz)
        return fd

    def var(self, v, ts=None):
        if ts is None: ts = len(self.times)-1
        return v.load(self, ts)
