// -*- c++ -*-

#ifndef MHD_H
#define MHD_H

#include <math.h>

#include "config.h"

static const int GX = GHOST_SIZE*(NX>1?1:0);
static const int GY = GHOST_SIZE*(NY>1?1:0);
static const int GZ = GHOST_SIZE*(NZ>1?1:0);

const int FACTOR = 1 << ORDER;  // factor for number of samples in each succesive level

#include "numtypes.h"

const unsigned long int M_TOTAL = ML * ( 1 - (unsigned long int)pow(FACTOR , L+1) ) / ( 1 - FACTOR );      // total number of samples (from all levels)

extern int NXl, NYl, NZl;

static int SIZE (int d) { return d==X?NXl:(d==Y?NYl:NZl); }
static bool HAVE_DIR (int d) { return (SIZE(d)>1); }

const int ACTIVE_DIRS = (NX>1) + (NY>1) + (NZ>1);
const int WORK_FACTOR = 1 << (ACTIVE_DIRS+1 - ORDER);

static int FLUX_NX (int d) { return d==X?NXl+1:NXl; }
static int FLUX_NY (int d) { return d==Y?NYl+1:NYl; }
static int FLUX_NZ (int d) { return d==Z?NZl+1:NZl; }

#define SIZE_CV 8
struct ConservedVars {
  real rho;
  real E;
  vector m;
  vector B;

  real norm() const
  {
    return fabs(rho)+m.norm()+fabs(E)+B.norm();
  }

  void zero()
  {
    rho = 0;
    m.zero();
    E = 0;
    B.zero();
  }
  
  void Exp() {
    rho = exp(rho);
    m.Exp();
    E   = exp(E);
    B.Exp();
  }
  
  void Log() {
    rho = log(rho);
    m.Log();
    E   = log(E);
    B.Log();
  }
  
  void nonnegative() {
    rho = rho>=0?rho:0;
    m.nonnegative();
    E   = E>=0?E:0;
    B.nonnegative();
  }
  
  void operator+=(const ConservedVars &a)
  {
    rho += a.rho;
    m += a.m;
    E += a.E;
    B += a.B;
  }

  void operator-=(const ConservedVars &a)
  {
    rho -= a.rho;
    m -= a.m;
    E -= a.E;
    B -= a.B;
  }
  
  void operator*=(const ConservedVars &a)
  {
    rho *= a.rho;
    m *= a.m;
    E *= a.E;
    B *= a.B;
  }

  void operator/=(const ConservedVars &a)
  {
    rho /= a.rho;
    m /= a.m;
    E /= a.E;
    B /= a.B;
  }
  
  void operator/=(const real M)
  {
    rho /= M;
    m /= M;
    E /= M;
    B /= M;
  }
  
  void operator*=(const real M)
  {
    rho *= M;
    m *= M;
    E *= M;
    B *= M;
  }
  
  #if COMPUTE_KURTOSIS
  void operator-=(const real M)
  {
    rho -= M;
    m -= M;
    E -= M;
    B -= M;
  }
  #endif
};

struct PrimitiveVars {
  real rho;
  real p;
  vector u;
  vector B;
  
  PrimitiveVars() {}
  PrimitiveVars(int i, int j, int k);

  void toConservedVars(ConservedVars &v) const {
    v.rho = rho;
    v.m = rho*u;
    v.E = p/(GAMMA-1) + 0.5*rho*u.sqr() + 0.5*B.sqr();
    v.B = B;
  }

  void zero() {
    rho = 0;
    p = 0;
    u.zero();
    B.zero();
  }
};

#define SIZE_XV 8
struct XtraVars {
  real p;
  real a;
  vector u;
  vector cf;
  
  void Exp() {
    p  = exp(p);
    a  = exp(a);
    u.Exp();
    cf.Exp();
  }
  
  void Log() {
    p  = log(p);
    a  = log(a);
    u.Log();
    cf.Log();
  }
  
  void nonnegative() {
    p  = p>=0?p:0;
    a  = a>=0?a:0;
    u.nonnegative();
    cf.nonnegative();
  }
  
  void operator+=(const XtraVars &x)
  {
    p += x.p;
    a += x.a;
    u += x.u;
    cf += x.cf;
  }
  
  void operator-=(const XtraVars &x)
  {
    p -= x.p;
    a -= x.a;
    u -= x.u;
    cf -= x.cf;
  }
  
  void operator*=(const XtraVars &x)
  {
    p *= x.p;
    a *= x.a;
    u *= x.u;
    cf *= x.cf;
  }
  
  void operator/=(const XtraVars &x)
  {
    p /= x.p;
    a /= x.a;
    u /= x.u;
    cf /= x.cf;
  }
  
  void operator/=(const real M)
  {
    p /= M;
    a /= M;
    u /= M;
    cf /= M;
  }
  
  void operator*=(const real M)
  {
    p *= M;
    a *= M;
    u *= M;
    cf *= M;
  }
};

struct AllVariables {
  ConservedVars U;
  XtraVars XU;
  
  AllVariables() {}
  AllVariables(ConservedVars &u, XtraVars &x): U(u), XU(x) {}
};

extern real dt, t, t_halt, halt_times[NSAVES+SAVE0], dx[NDIR];
extern volume<ConservedVars, 1> U;
extern volume<ConservedVars, 1> Flux[NDIR];
extern volume<ConservedVars, 1> Source;
extern volume<XtraVars, 1> Xtra;

// MLMC
// level dependent quantities
extern unsigned long int M[L+1];                // number of samples for each level
extern int N[NDIR][L+1];                        // number of cells for each level

extern int MULTI[NDIR][L+1];                    // number node groups working on the same domain
extern int NSAMPLERS[L+1];                      // number node groups working on the same level samples
extern int NSAMPLERSl;                          // number node groups working on the same level samples
extern int ACTIVE_SAMPLERSl;                    // number node groups working on the same downsampled level samples
extern int level_begin, level_end;              // levels to start and to end (might be different from 0 and L for multi:mpi)
extern unsigned long int partition_size;        // numbe of samples per one sampler

// main containers for mean and variance
extern volume <ConservedVars, 1> mean_U [NSAVES+SAVE0];
extern volume <XtraVars, 1>      mean_X [NSAVES+SAVE0];
#if COMPUTE_VARIANCE
extern volume <ConservedVars, 1> var_U  [NSAVES+SAVE0];
extern volume <XtraVars, 1>      var_X  [NSAVES+SAVE0];
#endif

// level-dependent containers for mean and variance
extern volume <ConservedVars, 1> mean_U_l [NSAVES+SAVE0];
extern volume <XtraVars, 1>      mean_X_l [NSAVES+SAVE0];
#if COMPUTE_VARIANCE
extern volume <ConservedVars, 1> var_U_l  [NSAVES+SAVE0];
extern volume <XtraVars, 1>      var_X_l  [NSAVES+SAVE0];
#endif

// temporary volumes for computational efficiency
#if COMPUTE_VARIANCE
extern volume <ConservedVars, 1> temp_U_l;
extern volume <XtraVars,      1> temp_X_l;
#endif

static inline ConservedVars operator+(const ConservedVars &a, const ConservedVars &b)
{
  ConservedVars v;
  v.rho = a.rho + b.rho;
  v.m = a.m + b.m;
  v.E = a.E + b.E;
  v.B = a.B + b.B;
  return v;
}

static inline ConservedVars operator-(const ConservedVars &a, const ConservedVars &b)
{
  ConservedVars v;
  v.rho = a.rho - b.rho;
  v.m = a.m - b.m;
  v.E = a.E - b.E;
  v.B = a.B - b.B;
  return v;
}

static inline ConservedVars operator*(real r, const ConservedVars &a)
{
  ConservedVars v;
  v.rho = r*a.rho;
  v.m = r*a.m;
  v.E = r*a.E;
  v.B = r*a.B;
  return v;
}

static inline ConservedVars operator*(const ConservedVars &a, real r) { return r*a; }

static inline ConservedVars operator/(const ConservedVars &a, real r)
{
  ConservedVars v;
  v.rho = a.rho/r;
  v.m = a.m/r;
  v.E = a.E/r;
  v.B = a.B/r;
  return v;
}

static inline ConservedVars operator*(const ConservedVars &a, const ConservedVars &b)
{
  ConservedVars v;
  v.rho = a.rho * b.rho;
  v.m = a.m || b.m;
  v.E = a.E * b.E;
  v.B = a.B || b.B;
  return v;
}

static inline XtraVars operator*(const XtraVars &xa, const XtraVars &xb)
{
  XtraVars x;
  x.p = xa.p * xb.p;
  x.a = xa.a * xb.a;
  x.u = xa.u || xb.u;
  x.cf = xa.cf || xb.cf;
  return x;
}

static inline XtraVars operator/(const XtraVars &xa, const real n)
{
  XtraVars x;
  x.p = xa.p / n;
  x.a = xa.a / n;
  x.u = xa.u / n;
  x.cf = xa.cf / n;
  return x;
}

static inline PrimitiveVars operator+(const PrimitiveVars &a, const PrimitiveVars &b)
{
  PrimitiveVars v;
  v.rho = a.rho + b.rho;
  v.u = a.u + b.u;
  v.p = a.p + b.p;
  v.B = a.B + b.B;
  return v;
}

static inline PrimitiveVars operator-(const PrimitiveVars &a, const PrimitiveVars &b)
{
  PrimitiveVars v;
  v.rho = a.rho - b.rho;
  v.u = a.u - b.u;
  v.p = a.p - b.p;
  v.B = a.B - b.B;
  return v;
}

static inline PrimitiveVars operator*(real r, const PrimitiveVars &a)
{
  PrimitiveVars v;
  v.rho = r*a.rho;
  v.u = r*a.u;
  v.p = r*a.p;
  v.B = r*a.B;
  return v;
}

static inline PrimitiveVars operator*(const PrimitiveVars &a, real r) { return r*a; }

static inline PrimitiveVars operator*(const PrimitiveVars &a, const PrimitiveVars &b)
{
  PrimitiveVars v;
  v.rho = a.rho * b.rho;
  v.u = a.u || b.u;
  v.p = a.p * b.p;
  v.B = a.B || b.B;
  return v;
}

inline PrimitiveVars::PrimitiveVars(int i, int j, int k)
{
  rho = U[i][j][k].rho;
  u = Xtra[i][j][k].u;
  p = Xtra[i][j][k].p;
  B = U[i][j][k].B;
}

void calc_point_flux(ConservedVars &F, const AllVariables &V, int d);
void time_step();
void sub_step(bool firsttime);
void step_init();
void bc_init();
void save_step();
real calc_time_step();
ConservedVars powell_calc(int i, int j, int k);
ConservedVars gravity_calc(int i, int j, int k);
void init_cell(PrimitiveVars &v, real x, real y, real z);
void set_ghost_x();
void set_ghost_y();
void set_ghost_z();
void calc_xtra(XtraVars &XU, const ConservedVars &U, int i, int j, int k);
void calc_all_vars(AllVariables &V, const PrimitiveVars &P);
void powell_source(vector &us, vector &Bs, AllVariables &lV, AllVariables &rV, int i, int j, int k, int d);

void mlmc_init ();
void level_init (int l);
void run_samples (int l, char action, bool compute);

/* Timers */
extern double timer;
void start_timer();
void stop_timer();

#include "multi.h"
#include "include.h"
#include "chaos.h"

#endif
